(**
 * Image tilize,
 *
 * contains functions to split images in smaller parts and use
 * pipelining mechanism to process images quicker
 *
 * Code should be compiled successfully under Debian Wheezy
 * with following line:
 *
 *   ocamlopt.opt -I +camlimages camlimages_core.cmxa unix.cmxa
 *      graphics.cmxa camlimages_graphics.cmxa
 *      camlimages_all.cmxa example.ml -o example
 *
 * It depends on: camlimages and standard ocaml library
 * (on Debian Wheezy: libcamlimages-ocaml-dev, libcamlimages-ocaml,
 * ocaml-native-compilers)
 *  
 * Code is distributed under the terms of the GNU General Public License 
 * (v2.0 or higher), described in file LICENSE.
 *                                                                    
 * (c) 2015 Andreas Romeyke
 * E-Mail: ocaml@andreas-romeyke.de
 *
 *)

(*
open Images
open OImages
open Rgb24
open Info
open Color
*)

type ipipeline_t = {
  width : int; (* original image *)
  height : int; (* original image *)
  overlap_width : int; (* how many pixel overlap? *)
  overlap_height : int; (* how many pixe overlap? *)
  tile_view_width : int; (* how many pixels viewable *)
  tile_view_height : int;
  tiles : OImages.rgb24 list; (* list of tile images *)
}

(** splits an image into an image pipeline
 * width = tile size (including overlap_width)
 * height = tile size (including overlap_height)
 *)
let pipelinize width height overlap_width overlap_height img =
  (*
  assert( img#width > width + overlap_width );
  assert( img#height > height + overlap_height );
  *)
  assert (overlap_width >= 0);
  assert (overlap_height >= 0);
  assert (width >= 0);
  assert (height >= 0);
  let tile_view_width = width - (2 * overlap_width) in
  let tile_view_height = height - (2 * overlap_height) in
  assert (tile_view_width > 0);
  assert (tile_view_height > 0);
  let subwidth = width in
  let subheight = height in
  Printf.printf "pipelinize (tileview w=%i h=%i, overlap w=%i h=%i):\n"
    tile_view_width tile_view_height overlap_width overlap_height;
  let w =
    ((1 + (img#width / tile_view_width)) * tile_view_width) + (2 * overlap_width)
  in
  let h =
    ((1 + (img#height / tile_view_height)) * tile_view_height)
    + (2 * overlap_height)
  in
  Printf.printf "\tw=%i h=%i\n" w h;
  let img' = new OImages.rgb24 w h in
  img#blit 0 0 img' overlap_width overlap_height img#width img#height;
  let rec _split (x : int) (y : int) tiles =
    if x > img'#width - 1 - (2 * overlap_width) then
      _split 0 (y + tile_view_height) tiles
    else if y > img'#height - 1 - (2 * overlap_height) then tiles
    else
      let (tmp : OImages.rgb24) = new OImages.rgb24 width height in
      Printf.printf "\tx0=%i y0=%i sw'=%i sh'=%i\n" x y subwidth subheight;
      img'#blit x y tmp 0 0 subwidth subheight;
      _split (x + tile_view_width) y (tmp :: tiles)
  in
  {
    width = img#width;
    height = img#height;
    overlap_width;
    overlap_height;
    tile_view_width = width - (2 * overlap_width);
    tile_view_height = height - (2 * overlap_height);
    tiles = List.rev (_split 0 0 []);
  }

(** join image pipeline to an image *)
let unpipelinize ipipeline =
  Printf.printf "unpipelinize: w=%i h=%i\n" ipipeline.width ipipeline.height;
  let (img : OImages.rgb24) =
    new OImages.rgb24 ipipeline.width ipipeline.height
  in
  let rec _join (x : int) (y : int) (lst : OImages.rgb24 list) =
    Printf.printf "\tx=%i y=%i ow=%i oh=%i\n" x y ipipeline.overlap_width
      ipipeline.overlap_height;
    match lst with
    | [] -> ()
    | lst when x >= ipipeline.width ->
        _join 0 (y + ipipeline.tile_view_height) lst
    | hd :: tl ->
        let subwidth = ipipeline.tile_view_width in
        let subheight = ipipeline.tile_view_height in
        let subwidth =
          if x + subwidth - 1 > img#width then img#width - x else subwidth
        in
        let subheight =
          if y + subheight - 1 > img#height then img#height - y else subheight
        in
        (*
      img#blit x y tmp 0 0 subwidth subheight ;
      *)
        Printf.printf
          "\t blitted x=%i y=%i from iox=%i ioy=%i with w=%i h=%i tw=%i th=%i\n"
          x y ipipeline.overlap_width ipipeline.overlap_height subwidth
          subheight ipipeline.tile_view_width ipipeline.tile_view_height;
        hd#blit ipipeline.overlap_width ipipeline.overlap_height img x y
          subwidth subheight;
        _join (x + ipipeline.tile_view_width) y tl
  in
  _join 0 0 ipipeline.tiles;
  img

let pipeline_map f pipeline =
  {
    width = pipeline.width;
    height = pipeline.height;
    overlap_width = pipeline.overlap_width;
    overlap_height = pipeline.overlap_height;
    tile_view_width = pipeline.tile_view_width;
    tile_view_height = pipeline.tile_view_height;
    tiles = List.map f pipeline.tiles;
  }
